import unittest
from unittest import mock

from flumine.baseflumine import (
    BaseFlumine,
    FlumineException,
    MaxTransactionCount,
    SimulatedMiddleware,
)
from flumine.clients import ExchangeType
from flumine.exceptions import ClientError


class BaseFlumineTest(unittest.TestCase):
    def setUp(self):
        self.mock_client = mock.Mock(EXCHANGE=ExchangeType.BETFAIR, paper_trade=False)
        self.base_flumine = BaseFlumine(self.mock_client)

    def test_init(self):
        self.assertFalse(self.base_flumine.SIMULATED)
        self.assertFalse(self.base_flumine._running)
        self.assertEqual(self.base_flumine._market_middleware, [])
        self.assertEqual(self.base_flumine._logging_controls, [])
        self.assertEqual(len(self.base_flumine.trading_controls), 3)
        self.assertEqual(self.base_flumine._workers, [])

    @mock.patch("flumine.baseflumine.SimulatedMiddleware")
    @mock.patch("flumine.baseflumine.BaseFlumine.add_market_middleware")
    def test_init_simulated(self, mock_add_market_middleware, mock_SimulatedMiddleware):
        BaseFlumine.SIMULATED = True
        mock_client = mock.Mock(EXCHANGE=ExchangeType.SIMULATED, paper_trade=False)
        BaseFlumine(mock_client)
        mock_add_market_middleware.assert_called_with(mock_SimulatedMiddleware())
        BaseFlumine.SIMULATED = False

    @mock.patch("flumine.baseflumine.SimulatedMiddleware")
    @mock.patch("flumine.baseflumine.BaseFlumine.add_market_middleware")
    def test_init_paper_trade(
        self, mock_add_market_middleware, mock_SimulatedMiddleware
    ):
        mock_client = mock.Mock(EXCHANGE=ExchangeType.BETFAIR, paper_trade=True)
        BaseFlumine(mock_client)
        mock_add_market_middleware.assert_called_with(mock_SimulatedMiddleware())

    def test_run(self):
        with self.assertRaises(NotImplementedError):
            self.base_flumine.run()

    @mock.patch("flumine.baseflumine.BaseFlumine.add_market_middleware")
    @mock.patch("flumine.baseflumine.BaseFlumine.add_client_control")
    def test_add_client(self, mock_add_client_control, mock_add_market_middleware):
        mock_clients = mock.Mock()
        self.base_flumine.clients = mock_clients
        mock_streams = mock.Mock()
        self.base_flumine.streams = mock_streams
        mock_client = mock.Mock()
        self.base_flumine.add_client(mock_client)
        mock_clients.add_client.assert_called_with(mock_client)
        mock_streams.add_client.assert_called_with(mock_client)
        mock_client.add_execution.assert_called_with(self.base_flumine)
        mock_add_market_middleware.assert_called()
        mock_add_client_control.assert_called_with(mock_client, MaxTransactionCount)

    @mock.patch("flumine.baseflumine.BaseFlumine.add_market_middleware")
    @mock.patch("flumine.baseflumine.BaseFlumine.add_client_control")
    def test_add_client_with_middleware(
        self, mock_add_client_control, mock_add_market_middleware
    ):
        self.base_flumine._market_middleware.append(SimulatedMiddleware())
        mock_clients = mock.Mock()
        self.base_flumine.clients = mock_clients
        mock_streams = mock.Mock()
        self.base_flumine.streams = mock_streams
        mock_client = mock.Mock()
        self.base_flumine.add_client(mock_client)
        mock_clients.add_client.assert_called_with(mock_client)
        mock_streams.add_client.assert_called_with(mock_client)
        mock_client.add_execution.assert_called_with(self.base_flumine)
        mock_add_market_middleware.assert_not_called()
        mock_add_client_control.assert_called_with(mock_client, MaxTransactionCount)

    @mock.patch("flumine.baseflumine.events")
    @mock.patch("flumine.baseflumine.BaseFlumine.log_control")
    def test_add_strategy(self, mock_log_control, mock_events):
        mock_streams = mock.Mock()
        self.base_flumine.streams = mock_streams
        mock_strategies = mock.Mock()
        self.base_flumine.strategies = mock_strategies
        mock_strategy = mock.Mock(market_filter={}, sports_data_filter=[])
        self.base_flumine.add_strategy(mock_strategy)
        mock_streams.assert_called_with(mock_strategy)
        mock_strategies.assert_called_with(mock_strategy, self.base_flumine.clients)
        mock_log_control.assert_called_with(mock_events.StrategyEvent(mock_strategy))

    def test_add_worker(self):
        mock_worker = mock.Mock()
        self.base_flumine.add_worker(mock_worker)
        self.assertEqual(len(self.base_flumine._workers), 1)

    def test_add_client_control(self):
        self.mock_client.trading_controls = []
        mock_control = mock.Mock()
        self.base_flumine.add_client_control(self.mock_client, mock_control)
        self.assertEqual(
            self.mock_client.trading_controls,
            [mock_control(self.base_flumine, self.mock_client)],
        )

    def test_add_trading_control(self):
        mock_control = mock.Mock()
        self.base_flumine.add_trading_control(mock_control)
        self.assertEqual(len(self.base_flumine.trading_controls), 4)

    def test_add_market_middleware(self):
        mock_middleware = mock.Mock()
        self.base_flumine.add_market_middleware(mock_middleware)
        self.assertEqual(len(self.base_flumine._market_middleware), 1)

    def test_add_logging_control(self):
        mock_control = mock.Mock()
        self.base_flumine.add_logging_control(mock_control)
        self.assertEqual(len(self.base_flumine._logging_controls), 1)

    def test_log_control(self):
        mock_control = mock.Mock()
        self.base_flumine._logging_controls.append(mock_control)
        mock_event = mock.Mock()
        self.base_flumine.log_control(mock_event)
        mock_control.logging_queue.put.assert_called_with(mock_event)

    def test__add_default_workers(self):
        self.base_flumine._add_default_workers()
        self.assertEqual(len(self.base_flumine._workers), 0)

    def test__process_market_books(self):
        mock_event = mock.Mock()
        mock_market_book = mock.Mock(publish_time_epoch=123)
        mock_market_book.runners = []
        mock_event.event = [mock_market_book]
        self.base_flumine._process_market_books(mock_event)

    @mock.patch("flumine.baseflumine.utils.call_strategy_error_handling")
    def test__process_sports_data(self, mock_call_strategy_error_handling):
        mock_market = mock.Mock()
        self.base_flumine.markets._markets = {"1.1": mock_market}
        mock_strategy_one = mock.Mock(stream_ids=[123])
        mock_strategy_two = mock.Mock(stream_ids=[])
        self.base_flumine.strategies = [
            mock_strategy_one,
            mock_strategy_two,
        ]
        mock_sports_data = mock.Mock(streaming_unique_id=123, market_id="1.1")
        mock_event = mock.Mock(event=[mock_sports_data])
        self.base_flumine._process_sports_data(mock_event)
        mock_call_strategy_error_handling.assert_has_calls(
            [
                mock.call(
                    mock_strategy_one.process_sports_data,
                    mock_market,
                    mock_sports_data,
                )
            ]
        )

    def test_process_order_package(self):
        mock_order_package = mock.Mock()
        self.base_flumine.process_order_package(mock_order_package)
        mock_order_package.client.execution.handler.assert_called_with(
            mock_order_package
        )

    @mock.patch("flumine.baseflumine.Market")
    def test__add_market(self, mock_market):
        mock_middleware = mock.Mock()
        self.base_flumine._market_middleware = [mock_middleware]
        mock_market_book = mock.Mock()
        self.assertEqual(
            self.base_flumine._add_market("1.234", mock_market_book), mock_market()
        )
        self.assertEqual(len(self.base_flumine.markets._markets), 1)
        mock_middleware.add_market.assert_called_with(mock_market())

    @mock.patch("flumine.baseflumine.BaseFlumine.info")
    def test__remove_market(self, _):
        mock_strategy = mock.Mock()
        self.base_flumine.strategies = [mock_strategy]
        mock_markets = mock.Mock()
        self.base_flumine.markets = mock_markets
        mock_middleware = mock.Mock()
        self.base_flumine._market_middleware = [mock_middleware]
        mock_market = mock.Mock()
        self.base_flumine._remove_market(mock_market)
        mock_markets.remove_market.assert_called_with(mock_market.market_id)
        mock_middleware.remove_market.assert_called_with(mock_market)
        mock_strategy.remove_market.assert_called_with(mock_market.market_id)

    @mock.patch("flumine.baseflumine.BaseFlumine.info")
    def test__remove_market_no_clear(self, _):
        mock_strategy = mock.Mock()
        self.base_flumine.strategies = [mock_strategy]
        mock_markets = mock.Mock()
        self.base_flumine.markets = mock_markets
        mock_middleware = mock.Mock()
        self.base_flumine._market_middleware = [mock_middleware]
        mock_market = mock.Mock()
        self.base_flumine._remove_market(mock_market, clear=False)
        mock_markets.remove_market.assert_not_called()
        mock_middleware.remove_market.assert_called_with(mock_market)
        mock_strategy.remove_market.assert_called_with(mock_market.market_id)

    @mock.patch("flumine.baseflumine.utils.call_process_raw_data")
    @mock.patch("flumine.baseflumine.BaseFlumine._add_market")
    def test__process_raw_data(self, mock__add_market, mock_call_process_raw_data):
        mock_strategy = mock.Mock(stream_ids=[12])
        self.base_flumine.strategies = [mock_strategy]
        mock_event = mock.Mock()
        mock_event.event = (12, "AAA", 12345, [{"id": "1.23"}])
        self.base_flumine._process_raw_data(mock_event)
        mock__add_market.assert_called_with("1.23", None)
        mock_call_process_raw_data.assert_called_with(
            mock_strategy, "AAA", 12345, {"id": "1.23"}
        )

    @mock.patch("flumine.baseflumine.events")
    @mock.patch("flumine.baseflumine.BaseFlumine._add_market")
    def test__process_raw_data_closed(self, mock__add_market, mock_events):
        mock_queue = mock.Mock()
        self.base_flumine.handler_queue = mock_queue
        mock_event = mock.Mock()
        mock_event.event = (
            12,
            "AAA",
            12345,
            [{"id": "1.23", "marketDefinition": {"status": "CLOSED"}}],
        )
        self.base_flumine._process_raw_data(mock_event)
        mock__add_market.assert_called_with("1.23", None)
        mock_queue.put.assert_called_with(mock_events.CloseMarketEvent())
        self.assertEqual(mock_event.event[3][0]["_stream_id"], mock_event.event[0])

    @mock.patch("flumine.baseflumine.BaseFlumine._add_market")
    def test__process_raw_data_no_id(self, mock__add_market):
        mock_event = mock.Mock()
        mock_event.event = (12, "AAA", 12345, [{"mid": "1.23"}])
        self.base_flumine._process_raw_data(mock_event)
        mock__add_market.assert_not_called()

    @mock.patch("flumine.baseflumine.events")
    @mock.patch("flumine.baseflumine.BaseFlumine.log_control")
    def test__process_market_catalogues(self, mock_log_control, mock_events):
        mock_market = mock.Mock()
        mock_market.market_catalogue = None
        mock_markets = mock.Mock()
        mock_markets.markets = {"1.23": mock_market}
        self.base_flumine.markets = mock_markets
        mock_market_catalogue = mock.Mock()
        mock_market_catalogue.market_id = "1.23"
        mock_event = mock.Mock()
        mock_event.event = [mock_market_catalogue]
        self.base_flumine._process_market_catalogues(mock_event)
        self.assertEqual(mock_market.market_catalogue, mock_market_catalogue)
        mock_log_control.assert_called_with(mock_events.MarketEvent(mock_market))
        self.assertFalse(mock_market.update_market_catalogue)

    @mock.patch("flumine.baseflumine.utils.call_process_orders_error_handling")
    @mock.patch("flumine.baseflumine.process_current_orders")
    def test__process_current_orders(
        self, mock_process_current_orders, mock_call_process_orders_error_handling
    ):
        mock_order = mock.Mock(complete=True)
        mock_market = mock.Mock(closed=False)
        mock_market.blotter.active = True
        mock_market.blotter.strategy_orders.return_value = [mock_order]
        self.base_flumine.markets = [mock_market]
        mock_strategy = mock.Mock()
        self.base_flumine.strategies = [mock_strategy]
        mock_current_orders = mock.Mock(orders=[mock_order])
        mock_event = mock.Mock(event=[mock_current_orders])
        self.base_flumine._process_current_orders(mock_event)
        mock_process_current_orders.assert_called_with(
            self.base_flumine.markets,
            self.base_flumine.strategies,
            mock_event,
            self.base_flumine.log_control,
            self.base_flumine._add_market,
        )
        mock_market.blotter.strategy_orders.assert_called_with(mock_strategy)
        mock_call_process_orders_error_handling.assert_called_with(
            mock_strategy, mock_market, [mock_order]
        )

    @mock.patch("flumine.baseflumine.process_current_orders")
    def test__process_current_orders_no_event(self, mock_process_current_orders):
        mock_event = mock.Mock(event=[])
        self.base_flumine._process_current_orders(mock_event)
        mock_process_current_orders.assert_not_called()

    def test__process_custom_event(self):
        mock_market = mock.Mock()
        self.base_flumine.markets = [mock_market]
        mock_event = mock.Mock()
        self.base_flumine._process_custom_event(mock_event)
        mock_event.callback.assert_called_with(self.base_flumine, mock_event)

    def test__process_custom_event_flu_error(self):
        mock_market = mock.Mock()
        self.base_flumine.markets = [mock_market]
        mock_event = mock.Mock()
        mock_event.callback.side_effect = FlumineException()
        self.base_flumine._process_custom_event(mock_event)
        mock_event.callback.assert_called_with(self.base_flumine, mock_event)

    def test__process_custom_event_error(self):
        mock_market = mock.Mock()
        self.base_flumine.markets = [mock_market]
        mock_event = mock.Mock()
        mock_event.callback.side_effect = ValueError()
        self.base_flumine._process_custom_event(mock_event)
        mock_event.callback.assert_called_with(self.base_flumine, mock_event)

    @mock.patch("flumine.baseflumine.config")
    def test__process_custom_event_error_raise(self, mock_config):
        mock_config.raise_errors = True
        mock_market = mock.Mock()
        self.base_flumine.markets = [mock_market]
        mock_event = mock.Mock()
        mock_event.callback.side_effect = ValueError()
        with self.assertRaises(ValueError):
            self.base_flumine._process_custom_event(mock_event)

    @mock.patch("flumine.baseflumine.BaseFlumine.info")
    @mock.patch("flumine.baseflumine.BaseFlumine.log_control")
    def test__process_close_market(self, mock_log_control, mock_info):
        mock_strategy = mock.Mock()
        mock_strategy.stream_ids = [1, 2, 3]
        self.base_flumine.strategies = [mock_strategy]
        mock_market = mock.Mock(closed=False, elapsed_seconds_closed=None)
        self.base_flumine.markets._markets = {"1.23": mock_market}
        mock_event = mock.Mock()
        mock_market_book = mock.Mock(market_id="1.23", streaming_unique_id=2)
        mock_event.event = mock_market_book
        self.base_flumine._process_close_market(mock_event)
        mock_market.close_market.assert_called_with()
        mock_market.blotter.process_closed_market.assert_called_with(mock_market_book)
        mock_strategy.process_closed_market.assert_called_with(
            mock_market, mock_market_book
        )
        mock_log_control.assert_called_with(mock_event)
        mock_market.assert_called_with(mock_market_book)

    @mock.patch("flumine.baseflumine.BaseFlumine.info")
    @mock.patch("flumine.baseflumine.BaseFlumine.log_control")
    def test__process_close_market_datum(self, mock_log_control, mock_info):
        mock_strategy = mock.Mock()
        mock_strategy.stream_ids = [1, 2, 3]
        self.base_flumine.strategies = [mock_strategy]
        mock_market = mock.Mock(closed=False, elapsed_seconds_closed=None)
        self.base_flumine.markets._markets = {"1.23": mock_market}
        mock_event = mock.Mock()
        mock_market_book = {"id": "1.23", "_stream_id": 2}
        mock_event.event = mock_market_book
        self.base_flumine._process_close_market(mock_event)
        mock_market.close_market.assert_called_with()
        mock_market.blotter.process_closed_market.assert_not_called()
        mock_strategy.process_closed_market.assert_called_with(
            mock_market, mock_market_book
        )
        mock_log_control.assert_called_with(mock_event)
        mock_market.assert_not_called()

    @mock.patch("flumine.baseflumine.BaseFlumine.info")
    def test__process_close_market_no_market(self, mock_info):
        mock_market = mock.Mock(closed=False, elapsed_seconds_closed=None)
        mock_market.market_book.streaming_unique_id = 2
        self.base_flumine.markets._markets = {"1.23": mock_market}
        mock_event = mock.Mock()
        mock_market_book = mock.Mock(market_id="1.45")
        mock_event.event = mock_market_book
        self.base_flumine._process_close_market(mock_event)
        mock_market.close_market.assert_not_called()

    @mock.patch("flumine.baseflumine.BaseFlumine.info")
    @mock.patch("flumine.baseflumine.BaseFlumine.log_control")
    def test__process_close_market_closed(self, mock_log_control, mock_info):
        mock_strategy = mock.Mock()
        mock_strategy.stream_ids = [1, 2, 3]
        self.base_flumine.strategies = [mock_strategy]
        mock_market = mock.Mock(
            market_id="1.23", event_id="1", closed=False, elapsed_seconds_closed=None
        )
        mock_market.market_book.streaming_unique_id = 2
        markets = [
            mock_market,
            mock.Mock(
                market_id="4.56", event_id="1", closed=True, elapsed_seconds_closed=25
            ),
            mock.Mock(
                market_id="7.89", event_id="1", closed=True, elapsed_seconds_closed=3601
            ),
            mock.Mock(
                market_id="1.01",
                event_id="2",
                closed=False,
                elapsed_seconds_closed=3601,
            ),
        ]
        for market in markets:
            self.base_flumine.markets.add_market(market.market_id, market)
        mock_market_book = mock.Mock(market_id="1.23")
        mock_event = mock.Mock(event=mock_market_book)
        self.base_flumine._process_close_market(mock_event)

        self.assertEqual(len(self.base_flumine.markets._markets), 3)
        self.assertEqual(len(self.base_flumine.markets.events), 2)

    @mock.patch("flumine.baseflumine.BaseFlumine._process_cleared_markets")
    @mock.patch("flumine.baseflumine.events")
    @mock.patch("flumine.baseflumine.BaseFlumine._process_cleared_orders")
    @mock.patch("flumine.baseflumine.BaseFlumine.info")
    @mock.patch("flumine.baseflumine.BaseFlumine.log_control")
    def test__process_close_market_closed_paper(
        self,
        mock_log_control,
        mock_info,
        mock__process_cleared_orders,
        mock_events,
        mock__process_cleared_markets,
    ):
        self.mock_client.paper_trade = True
        mock_strategy = mock.Mock()
        mock_strategy.stream_ids = [1, 2, 3]
        self.base_flumine.strategies = [mock_strategy]
        mock_market = mock.Mock(closed=False, elapsed_seconds_closed=None)
        mock_market.market_book.streaming_unique_id = 2
        mock_market.cleared.return_value = {}
        self.base_flumine.markets._markets = {
            "1.23": mock_market,
            "4.56": mock.Mock(market_id="4.56", closed=True, elapsed_seconds_closed=25),
            "7.89": mock.Mock(
                market_id="7.89", closed=True, elapsed_seconds_closed=3601
            ),
            "1.01": mock.Mock(
                market_id="1.01", closed=False, elapsed_seconds_closed=3601
            ),
        }
        mock_event = mock.Mock()
        mock_market_book = mock.Mock(market_id="1.23")
        mock_event.event = mock_market_book
        self.base_flumine._process_close_market(mock_event)
        self.assertEqual(len(self.base_flumine.markets._markets), 4)
        mock__process_cleared_orders.assert_called_with(
            mock_events.ClearedOrdersEvent()
        )
        mock_market.cleared.assert_called_with(self.mock_client)
        mock__process_cleared_markets.assert_called_with(
            mock_events.ClearedMarketsEvent()
        )

    @mock.patch("flumine.baseflumine.events")
    @mock.patch("flumine.baseflumine.BaseFlumine.log_control")
    @mock.patch("flumine.baseflumine.BaseFlumine.info")
    def test__process_cleared_orders(self, mock_info, mock_log_control, mock_events):
        mock_market = mock.Mock()
        mock_market.blotter.process_cleared_orders.return_value = []
        mock_markets = mock.Mock()
        mock_markets.markets = {"1.23": mock_market}
        self.base_flumine.markets = mock_markets
        mock_event = mock.Mock()
        mock_event.event.market_id = "1.23"
        mock_event.event.orders = []
        self.base_flumine._process_cleared_orders(mock_event)
        mock_market.blotter.process_cleared_orders.assert_called_with(mock_event.event)
        mock_log_control.assert_called_with(mock_events.ClearedOrdersMetaEvent())

    @mock.patch("flumine.baseflumine.BaseFlumine.info")
    def test__process_cleared_orders_no_market(self, mock_info):
        mock_market = mock.Mock()
        mock_markets = mock.Mock()
        mock_markets.markets = {"1.23": mock_market}
        self.base_flumine.markets = mock_markets
        mock_event = mock.Mock()
        mock_event.event.market_id = "1.24"
        mock_event.event.orders = []
        self.base_flumine._process_cleared_orders(mock_event)
        mock_market.blotter.process_cleared_orders.assert_not_called()

    @mock.patch("flumine.baseflumine.BaseFlumine.log_control")
    def test__process_cleared_markets(self, mock_log_control):
        mock_event = mock.Mock()
        mock_event.event.orders = []
        self.base_flumine._process_cleared_markets(mock_event)
        mock_log_control.assert_called_with(mock_event)

    def test__process_end_flumine(self):
        mock_strategies = mock.Mock()
        self.base_flumine.strategies = mock_strategies
        self.base_flumine._process_end_flumine()
        mock_strategies.finish.assert_called_with(self.base_flumine)

    def test_info(self):
        self.assertTrue(self.base_flumine.info)

    def test_enter_no_clients(self):
        self.base_flumine.clients._clients = []
        with self.assertRaises(ClientError):
            with self.base_flumine:
                pass

    @mock.patch("flumine.baseflumine.BaseFlumine._process_end_flumine")
    @mock.patch("flumine.baseflumine.events")
    @mock.patch("flumine.baseflumine.BaseFlumine.log_control")
    def test_enter_exit(self, mock_log_control, mock_events, mock__process_end_flumine):
        control = mock.Mock()
        self.base_flumine._logging_controls = [control]
        self.base_flumine.simulated_execution = mock.Mock()
        self.base_flumine.betfair_execution = mock.Mock()
        with self.base_flumine:
            self.assertTrue(self.base_flumine._running)
            self.mock_client.login.assert_called_with()
            mock_log_control.assert_called_with(mock_events.ConfigEvent(None))

        self.assertFalse(self.base_flumine._running)
        mock__process_end_flumine.assert_called_with()
        self.mock_client.logout.assert_called_with()
        self.base_flumine.simulated_execution.shutdown.assert_called_with()
        self.base_flumine.betfair_execution.shutdown.assert_called_with()
        control.start.assert_called_with()
        mock_log_control.assert_called_with(mock_events.TerminationEvent(None))



import unittest
from unittest import mock

from flumine.simulation import simulatedorder
from flumine.order.ordertype import OrderTypes


class SimulatedOrderTest(unittest.TestCase):
    def setUp(self) -> None:
        self.mock_order_type = mock.Mock(
            price=12, size=2.00, ORDER_TYPE=OrderTypes.LIMIT
        )
        mock_client = mock.Mock(
            paper_trade=False, min_bsp_liability=10, simulated_full_match=False
        )
        mock_trade = mock.Mock()
        self.mock_order = mock.Mock(
            selection_id=1234,
            handicap=1,
            side="BACK",
            order_type=self.mock_order_type,
            trade=mock_trade,
            number_of_dead_heat_winners=None,
            client=mock_client,
        )
        self.simulated = simulatedorder.SimulatedOrder(self.mock_order)

    def test_init(self):
        self.assertEqual(self.simulated.order, self.mock_order)
        self.assertEqual(self.simulated.matched, [])
        self.assertEqual(self.simulated.size_cancelled, 0)
        self.assertEqual(self.simulated.size_lapsed, 0)
        self.assertEqual(self.simulated.size_voided, 0)
        self.assertEqual(self.simulated._piq, 0)
        self.assertIsNone(self.simulated.market_version)
        self.assertFalse(self.simulated._bsp_reconciled)

    @mock.patch("flumine.simulation.simulatedorder.SimulatedOrder._get_runner")
    @mock.patch(
        "flumine.simulation.simulatedorder.SimulatedOrder.take_sp", return_value=True
    )
    @mock.patch("flumine.simulation.simulatedorder.SimulatedOrder._process_traded")
    @mock.patch("flumine.simulation.simulatedorder.SimulatedOrder._process_sp")
    def test_call_process_traded(
        self, mock__process_sp, mock__process_traded, _, mock__get_runner
    ):
        mock_market_book = mock.Mock(bsp_reconciled=False)
        traded = {1: 2}
        self.simulated(mock_market_book, traded)
        mock__process_sp.assert_not_called()
        mock__process_traded.assert_called_with(
            mock_market_book.publish_time_epoch, traded
        )

    @mock.patch("flumine.simulation.simulatedorder.SimulatedOrder._get_runner")
    @mock.patch(
        "flumine.simulation.simulatedorder.SimulatedOrder.take_sp", return_value=True
    )
    @mock.patch("flumine.simulation.simulatedorder.SimulatedOrder._process_traded")
    @mock.patch("flumine.simulation.simulatedorder.SimulatedOrder._process_sp")
    def test_call_process_sp(
        self, mock__process_sp, mock__process_traded, _, mock__get_runner
    ):
        mock_market_book = mock.Mock()
        mock_market_book.bsp_reconciled = True
        mock_runner_analytics = mock.Mock()
        self.simulated(mock_market_book, mock_runner_analytics)
        mock__process_sp.assert_called_with(
            mock_market_book.publish_time_epoch, mock__get_runner()
        )
        mock__process_traded.assert_not_called()

    @mock.patch("flumine.simulation.simulatedorder.SimulatedOrder._get_runner")
    @mock.patch(
        "flumine.simulation.simulatedorder.SimulatedOrder.take_sp", return_value=True
    )
    @mock.patch("flumine.simulation.simulatedorder.SimulatedOrder._process_traded")
    @mock.patch("flumine.simulation.simulatedorder.SimulatedOrder._process_sp")
    def test_call_market_version(
        self, mock__process_sp, mock__process_traded, _, mock__get_runner
    ):
        self.simulated.market_version = 123
        self.simulated.order.order_type.persistence_type = "LAPSE"
        mock_market_book = mock.Mock(
            bsp_reconciled=False, version=124, status="SUSPENDED"
        )
        mock_runner_analytics = mock.Mock()
        self.simulated(mock_market_book, mock_runner_analytics)
        self.assertEqual(self.simulated.size_lapsed, 2.0)
        self.assertEqual(self.simulated.size_remaining, 0.0)
        mock__process_sp.assert_not_called()
        mock__process_traded.assert_not_called()

    @mock.patch("flumine.simulation.simulatedorder.SimulatedOrder._get_runner")
    @mock.patch(
        "flumine.simulation.simulatedorder.SimulatedOrder.take_sp", return_value=True
    )
    @mock.patch("flumine.simulation.simulatedorder.SimulatedOrder._process_traded")
    @mock.patch("flumine.simulation.simulatedorder.SimulatedOrder._process_sp")
    def test_call_not_re(
        self, mock__process_sp, mock__process_traded, _, mock__get_runner
    ):
        mock_market_book = mock.Mock()
        mock_market_book.bsp_reconciled = False
        self.simulated.order.order_type.ORDER_TYPE = OrderTypes.MARKET_ON_CLOSE
        self.simulated(mock_market_book, {})
        mock__process_sp.assert_not_called()
        mock__process_traded.assert_not_called()

        mock_market_book.bsp_reconciled = True
        self.simulated(mock_market_book, {})
        mock__process_sp.assert_called_with(
            mock_market_book.publish_time_epoch, mock__get_runner()
        )
        mock__process_traded.assert_not_called()

    @mock.patch("flumine.simulation.simulatedorder.SimulatedOrder._get_runner")
    @mock.patch(
        "flumine.simulation.simulatedorder.SimulatedOrder.take_sp", return_value=True
    )
    @mock.patch("flumine.simulation.simulatedorder.SimulatedOrder._process_traded")
    @mock.patch("flumine.simulation.simulatedorder.SimulatedOrder._process_sp")
    def test_call_bsp_reconciled(
        self, mock__process_sp, mock__process_traded, _, mock__get_runner
    ):
        self.simulated._bsp_reconciled = True
        mock_market_book = mock.Mock()
        mock_market_book.bsp_reconciled = True
        self.simulated.order.order_type.ORDER_TYPE = OrderTypes.MARKET_ON_CLOSE
        self.simulated(mock_market_book, {})
        mock__process_sp.assert_not_called()
        mock__process_traded.assert_not_called()

    @mock.patch("flumine.simulation.simulatedorder.SimulatedOrder._get_runner")
    def test_place_limit_back(self, mock__get_runner):
        mock_client = mock.Mock(best_price_execution=True)
        mock_order_package = mock.Mock(client=mock_client, market_version=None)
        mock_market_book = mock.Mock(status="OPEN")
        mock_runner = mock.Mock()
        mock_runner.ex.available_to_back = [{"price": 12, "size": 120}]
        mock_runner.ex.available_to_lay = [{"price": 13, "size": 120}]
        mock__get_runner.return_value = mock_runner
        resp = self.simulated.place(mock_order_package, mock_market_book, {}, 1)
        self.assertEqual(self.simulated.market_version, mock_market_book.version)
        self.assertEqual(resp.average_price_matched, 12)
        self.assertEqual(resp.size_matched, 2)
        self.assertEqual(
            self.simulated.matched, [[mock_market_book.publish_time_epoch, 12, 2]]
        )

    @mock.patch("flumine.simulation.simulatedorder.SimulatedOrder._get_runner")
    def test_place_limit_back_target_size(self, mock__get_runner):
        self.mock_order.order_type.size = None
        mock_client = mock.Mock(best_price_execution=True)
        mock_order_package = mock.Mock(client=mock_client, market_version=None)
        mock_market_book = mock.Mock(status="OPEN")
        mock_runner = mock.Mock()
        mock_runner.ex.available_to_back = [{"price": 12, "size": 120}]
        mock_runner.ex.available_to_lay = [{"price": 13, "size": 120}]
        mock__get_runner.return_value = mock_runner
        with self.assertRaises(NotImplementedError):
            self.simulated.place(mock_order_package, mock_market_book, {}, 1)

    @mock.patch("flumine.simulation.simulatedorder.SimulatedOrder._get_runner")
    def test_place_market_status(self, mock__get_runner):
        mock_client = mock.Mock(best_price_execution=False)
        mock_order_package = mock.Mock(client=mock_client, market_version=None)
        mock_market_book = mock.Mock(status="SUSPENDED")
        resp = self.simulated.place(mock_order_package, mock_market_book, {}, 1)
        self.assertEqual(resp.status, "FAILURE")
        self.assertEqual(resp.error_code, "ERROR_IN_ORDER")
        self.assertEqual(self.simulated.matched, [])
        self.assertEqual(self.simulated.size_voided, 2)

    @mock.patch("flumine.simulation.simulatedorder.SimulatedOrder._get_runner")
    def test_place_market_version(self, mock__get_runner):
        mock_client = mock.Mock(best_price_execution=False)
        mock_order_package = mock.Mock(
            client=mock_client, market_version={"version": 123}
        )
        mock_market_book = mock.Mock(status="OPEN", version=124)
        resp = self.simulated.place(mock_order_package, mock_market_book, {}, 1)
        self.assertEqual(resp.status, "FAILURE")
        self.assertEqual(resp.error_code, "ERROR_IN_ORDER")
        self.assertEqual(self.simulated.matched, [])
        self.assertEqual(self.simulated.size_voided, 2)

    @mock.patch("flumine.simulation.simulatedorder.SimulatedOrder._get_runner")
    def test_place_limit_back_unmatched(self, mock__get_runner):
        mock_client = mock.Mock(best_price_execution=True)
        mock_order_package = mock.Mock(client=mock_client, market_version=None)
        mock_market_book = mock.Mock(status="OPEN")
        mock_runner = mock.Mock()
        mock_runner.ex.available_to_back = [{"price": 10, "size": 120}]
        mock_runner.ex.available_to_lay = [
            {"price": 10.5, "size": 120},
            {"price": 11.5, "size": 10},
            {"price": 12, "size": 22},
            {"price": 15, "size": 32},
        ]
        mock__get_runner.return_value = mock_runner
        resp = self.simulated.place(mock_order_package, mock_market_book, {}, 1)
        self.assertEqual(resp.average_price_matched, 0)
        self.assertEqual(resp.size_matched, 0)
        self.assertEqual(self.simulated.matched, [])
        self.assertEqual(self.simulated._piq, 22)

    @mock.patch("flumine.simulation.simulatedorder.SimulatedOrder._get_runner")
    def test_place_limit_back_bpe(self, mock__get_runner):
        mock_client = mock.Mock(best_price_execution=False)
        mock_order_package = mock.Mock(client=mock_client, market_version=None)
        mock_market_book = mock.Mock(status="OPEN")
        mock_runner = mock.Mock()
        mock_runner.ex.available_to_back = [{"price": 15, "size": 120}]
        mock_runner.ex.available_to_lay = [{"price": 16, "size": 120}]
        mock__get_runner.return_value = mock_runner
        resp = self.simulated.place(mock_order_package, mock_market_book, {}, 1)
        self.assertEqual(resp.status, "FAILURE")
        self.assertEqual(resp.error_code, "BET_LAPSED_PRICE_IMPROVEMENT_TOO_LARGE")
        self.assertEqual(self.simulated.matched, [])
        self.assertEqual(self.simulated.size_lapsed, 2)

    @mock.patch("flumine.simulation.simulatedorder.SimulatedOrder._get_runner")
    def test_place_limit_lay(self, mock__get_runner):
        mock_client = mock.Mock(best_price_execution=True)
        mock_order_package = mock.Mock(client=mock_client, market_version=None)
        self.simulated.order.side = "LAY"
        mock_market_book = mock.Mock(status="OPEN")
        mock_runner = mock.Mock()
        mock_runner.ex.available_to_back = [{"price": 11, "size": 120}]
        mock_runner.ex.available_to_lay = [{"price": 12, "size": 120}]
        mock__get_runner.return_value = mock_runner
        resp = self.simulated.place(mock_order_package, mock_market_book, {}, 1)
        self.assertEqual(resp.average_price_matched, 12)
        self.assertEqual(resp.size_matched, 2)
        self.assertEqual(
            self.simulated.matched, [[mock_market_book.publish_time_epoch, 12, 2]]
        )

    @mock.patch("flumine.simulation.simulatedorder.SimulatedOrder._get_runner")
    def test_place_limit_lay_unmatched(self, mock__get_runner):
        mock_client = mock.Mock(best_price_execution=True)
        mock_order_package = mock.Mock(client=mock_client, market_version=None)
        self.simulated.order.side = "LAY"
        mock_market_book = mock.Mock(status="OPEN")
        mock_runner = mock.Mock()
        mock_runner.ex.available_to_back = [
            {"price": 10.5, "size": 120},
            {"price": 11.5, "size": 10},
            {"price": 12, "size": 22},
            {"price": 14, "size": 32},
        ]
        mock_runner.ex.available_to_lay = [{"price": 15, "size": 32}]
        mock__get_runner.return_value = mock_runner
        resp = self.simulated.place(mock_order_package, mock_market_book, {}, 1)
        self.assertEqual(resp.average_price_matched, 0)
        self.assertEqual(resp.size_matched, 0)
        self.assertEqual(self.simulated.matched, [])
        self.assertEqual(self.simulated._piq, 22)

    @mock.patch("flumine.simulation.simulatedorder.SimulatedOrder._get_runner")
    def test_place_limit_lay_bpe(self, mock__get_runner):
        mock_client = mock.Mock(best_price_execution=False)
        mock_order_package = mock.Mock(client=mock_client, market_version=None)
        self.simulated.order.side = "LAY"
        mock_market_book = mock.Mock(status="OPEN")
        mock_runner = mock.Mock()
        mock_runner.ex.available_to_back = [{"price": 10, "size": 120}]
        mock_runner.ex.available_to_lay = [{"price": 10.5, "size": 120}]
        mock__get_runner.return_value = mock_runner
        resp = self.simulated.place(mock_order_package, mock_market_book, {}, 1)
        self.assertEqual(resp.status, "FAILURE")
        self.assertEqual(resp.error_code, "BET_LAPSED_PRICE_IMPROVEMENT_TOO_LARGE")
        self.assertEqual(self.simulated.matched, [])
        self.assertEqual(self.simulated.size_lapsed, 2)

    @mock.patch("flumine.simulation.simulatedorder.SimulatedOrder._get_runner")
    def test_place_limit_removed_runner(self, mock__get_runner):
        mock_client = mock.Mock(best_price_execution=False)
        mock_order_package = mock.Mock(client=mock_client, market_version=None)
        self.simulated.order.side = "BACK"
        mock_market_book = mock.Mock(status="OPEN")
        mock_runner = mock.Mock()
        mock_runner.ex.available_to_back = [{"price": 10, "size": 120}]
        mock_runner.ex.available_to_lay = [{"price": 10.5, "size": 120}]
        mock_runner.status = "REMOVED"
        mock__get_runner.return_value = mock_runner
        resp = self.simulated.place(mock_order_package, mock_market_book, {}, 1)
        self.assertEqual(resp.status, "FAILURE")
        self.assertEqual(resp.error_code, "RUNNER_REMOVED")
        self.assertEqual(self.simulated.matched, [])
        self.assertEqual(self.simulated.size_voided, 2)

    @mock.patch(
        "flumine.simulation.simulatedorder.SimulatedOrder._create_place_response"
    )
    @mock.patch("flumine.simulation.simulatedorder.SimulatedOrder._get_runner")
    def test_place_else(self, mock__get_runner, mock__create_place_response):
        mock__get_runner.status = "ACTIVE"
        mock_client = mock.Mock(best_price_execution=True)
        mock_order_package = mock.Mock(client=mock_client, market_version=None)
        self.simulated.order.order_type.ORDER_TYPE = OrderTypes.MARKET_ON_CLOSE
        mock_market_book = mock.Mock(status="OPEN")
        self.simulated.place(mock_order_package, mock_market_book, {}, 1)
        self.assertEqual(self.simulated.matched, [])
        mock__create_place_response.assert_called_with(1)

    @mock.patch("flumine.simulation.simulatedorder.SimulatedOrder._get_runner")
    def test_place_else_bsp_recon(self, mock__get_runner):
        mock__get_runner.status = "ACTIVE"
        mock_client = mock.Mock(best_price_execution=True)
        mock_order_package = mock.Mock(client=mock_client, market_version=None)
        self.simulated.order.order_type.ORDER_TYPE = OrderTypes.MARKET_ON_CLOSE
        mock_market_book = mock.Mock(status="OPEN", bsp_reconciled=True, inplay=False)
        resp = self.simulated.place(mock_order_package, mock_market_book, {}, 1)
        self.assertEqual(resp.status, "FAILURE")
        self.assertEqual(resp.error_code, "MARKET_NOT_OPEN_FOR_BSP_BETTING")
        self.assertEqual(self.simulated.matched, [])
        self.assertEqual(self.simulated.size_voided, 0)

    def test__create_place_response(self):
        resp = self.simulated._create_place_response(
            1234, "FAILURE", error_code="dubs of the mad skint and british"
        )
        self.assertEqual(resp.bet_id, "1234")
        self.assertEqual(resp.status, "FAILURE")
        self.assertEqual(resp.order_status, "EXECUTABLE")
        self.assertEqual(resp.error_code, "dubs of the mad skint and british")

    @mock.patch(
        "flumine.simulation.simulatedorder.SimulatedOrder.size_remaining",
        new_callable=mock.PropertyMock,
        return_value=0,
    )
    def test__create_place_response_complete(self, mock_size_remaining):
        resp = self.simulated._create_place_response(1234)
        self.assertEqual(resp.bet_id, "1234")
        self.assertEqual(resp.status, "SUCCESS")
        self.assertEqual(resp.order_status, "EXECUTION_COMPLETE")

    def test__create_place_response_full_match(self):
        self.mock_order.client.simulated_full_match = True
        resp = self.simulated._create_place_response(1234)
        self.assertEqual(resp.average_price_matched, 12)
        self.assertEqual(resp.size_matched, 2)
        self.assertEqual(self.simulated.matched, [[0, 12, 2.0]])
        self.assertEqual(self.simulated.size_matched, 2)
        self.assertEqual(self.simulated.average_price_matched, 12)
        self.assertEqual(resp.bet_id, "1234")
        self.assertEqual(resp.status, "SUCCESS")
        self.assertEqual(resp.order_status, "EXECUTION_COMPLETE")

    def test__create_place_response_failure_full_match(self):
        self.mock_order.client.simulated_full_match = True
        resp = self.simulated._create_place_response(None, status="FAILURE")
        self.assertEqual(resp.average_price_matched, 0)
        self.assertEqual(resp.size_matched, 0)
        self.assertEqual(self.simulated.matched, [])
        self.assertEqual(self.simulated.size_matched, 0)
        self.assertEqual(self.simulated.average_price_matched, 0)
        self.assertIsNone(resp.bet_id)
        self.assertEqual(resp.status, "FAILURE")
        self.assertEqual(resp.order_status, "EXECUTABLE")

    def test_cancel(self):
        self.simulated.order.update_data = {}
        mock_market_book = mock.Mock(status="OPEN")
        resp = self.simulated.cancel(mock_market_book)
        self.assertEqual(self.simulated.size_cancelled, 2.0)
        self.assertEqual(resp.status, "SUCCESS")
        self.assertEqual(resp.size_cancelled, 2.0)

    def test_cancel_market_status(self):
        mock_market_book = mock.Mock(status="SUSPENDED")
        resp = self.simulated.cancel(mock_market_book)
        self.assertEqual(resp.status, "FAILURE")
        self.assertEqual(resp.error_code, "ERROR_IN_ORDER")
        self.assertEqual(self.simulated.size_cancelled, 0)

    def test_cancel_reduction(self):
        self.simulated.order.update_data = {"size_reduction": 0.50}
        mock_market_book = mock.Mock(status="OPEN")
        resp = self.simulated.cancel(mock_market_book)
        self.assertEqual(self.simulated.size_cancelled, 0.50)
        self.assertEqual(self.simulated.size_remaining, 1.50)
        self.assertEqual(resp.status, "SUCCESS")
        self.assertEqual(resp.size_cancelled, 0.50)

    def test_cancel_reduction_multi(self):
        self.simulated.size_cancelled = 0.10
        self.simulated.order.update_data = {"size_reduction": 0.50}
        mock_market_book = mock.Mock(status="OPEN")
        resp = self.simulated.cancel(mock_market_book)
        self.assertEqual(self.simulated.size_cancelled, 0.60)
        self.assertEqual(self.simulated.size_remaining, 1.40)
        self.assertEqual(resp.status, "SUCCESS")
        self.assertEqual(resp.size_cancelled, 0.50)

    def test_cancel_reduction_greater_than(self):
        self.simulated.order.update_data = {"size_reduction": 64.0}
        mock_market_book = mock.Mock(status="OPEN")
        resp = self.simulated.cancel(mock_market_book)
        self.assertEqual(self.simulated.size_cancelled, 2.00)
        self.assertEqual(self.simulated.size_remaining, 0.00)
        self.assertEqual(resp.status, "SUCCESS")
        self.assertEqual(resp.size_cancelled, 2.00)

    def test_cancel_else(self):
        self.simulated.order.order_type.ORDER_TYPE = OrderTypes.MARKET_ON_CLOSE
        mock_market_book = mock.Mock(status="OPEN")
        resp = self.simulated.cancel(mock_market_book)
        self.assertEqual(resp.status, "FAILURE")
        self.assertEqual(resp.error_code, "BET_ACTION_ERROR")

    def test_update(self):
        mock_market_book = mock.Mock(status="OPEN")
        resp = self.simulated.update(
            mock_market_book, {"newPersistenceType": "PERSIST"}
        )
        self.assertEqual(resp.status, "SUCCESS")
        self.assertEqual(self.simulated.order.order_type.persistence_type, "PERSIST")

    def test_update_market_status(self):
        mock_market_book = mock.Mock(status="SUSPENDED")
        resp = self.simulated.update(
            mock_market_book, {"newPersistenceType": "PERSIST"}
        )
        self.assertEqual(resp.status, "FAILURE")
        self.assertEqual(resp.error_code, "ERROR_IN_ORDER")

    def test_update_market_p_enabled(self):
        mock_market_book = mock.Mock(status="OPEN")
        mock_market_book.market_definition.persistence_enabled = False
        resp = self.simulated.update(
            mock_market_book, {"newPersistenceType": "PERSIST"}
        )
        self.assertEqual(resp.status, "FAILURE")
        self.assertEqual(resp.error_code, "INVALID_PERSISTENCE_TYPE")

    def test_update_else(self):
        mock_market_book = mock.Mock(status="OPEN")
        self.simulated.order.order_type.ORDER_TYPE = OrderTypes.MARKET_ON_CLOSE
        resp = self.simulated.update(
            mock_market_book, {"newPersistenceType": "PERSIST"}
        )
        self.assertEqual(resp.status, "FAILURE")
        self.assertEqual(resp.error_code, "BET_ACTION_ERROR")

    def test__get_runner(self):
        mock_market_book = mock.Mock()
        mock_runner = mock.Mock(selection_id=1234, handicap=1)
        mock_market_book.runners = [mock_runner]
        self.assertEqual(
            self.simulated._get_runner(mock_market_book),
            mock_runner,
        )
        mock_runner = mock.Mock(selection_id=134, handicap=1)
        mock_market_book.runners = [mock_runner]
        self.assertIsNone(self.simulated._get_runner(mock_market_book))

    @mock.patch(
        "flumine.simulation.simulatedorder.SimulatedOrder.side",
        new_callable=mock.PropertyMock,
        return_value="BACK",
    )
    def test__process_price_matched_back(self, mock_side):
        self.simulated._process_price_matched(
            1234567, 12.0, 2.00, [{"price": 15, "size": 120}]
        )
        self.assertEqual(self.simulated.matched, [[1234567, 15, 2]])
        self.simulated.matched = []
        self.simulated._process_price_matched(
            1234568, 12.0, 2.00, [{"price": 15, "size": 1}, {"price": 12, "size": 1}]
        )
        self.assertEqual(self.simulated.matched, [[1234568, 15, 1], [1234568, 12, 1]])
        self.simulated.matched = []
        self.simulated._process_price_matched(
            1234569, 12.0, 2.00, [{"price": 15, "size": 1}, {"price": 12, "size": 0.5}]
        )
        self.assertEqual(self.simulated.matched, [[1234569, 15, 1], [1234569, 12, 0.5]])
        self.simulated.matched = []
        self.simulated._process_price_matched(
            1234570, 12.0, 2.00, [{"price": 15, "size": 1}, {"price": 11, "size": 0.5}]
        )
        self.assertEqual(self.simulated.matched, [[1234570, 15, 1]])

    @mock.patch(
        "flumine.simulation.simulatedorder.SimulatedOrder.side",
        new_callable=mock.PropertyMock,
        return_value="LAY",
    )
    def test__process_price_matched_lay(self, mock_side):
        self.simulated._process_price_matched(
            1234571, 3.0, 20.00, [{"price": 2.02, "size": 120}]
        )
        self.assertEqual(self.simulated.matched, [[1234571, 2.02, 20]])
        self.simulated.matched = []
        self.simulated._process_price_matched(
            1234571, 3.0, 20.00, [{"price": 2.02, "size": 1}, {"price": 3, "size": 20}]
        )
        self.assertEqual(self.simulated.matched, [[1234571, 2.02, 1], [1234571, 3, 19]])
        self.simulated.matched = []
        self.simulated._process_price_matched(
            1234571,
            3.0,
            20.00,
            [{"price": 2.02, "size": 1}, {"price": 2.9, "size": 0.5}],
        )
        self.assertEqual(
            self.simulated.matched, [[1234571, 2.02, 1], [1234571, 2.9, 0.5]]
        )
        self.simulated.matched = []
        self.simulated._process_price_matched(
            1234571, 3.0, 20.00, [{"price": 3, "size": 1}, {"price": 11, "size": 0.5}]
        )
        self.assertEqual(self.simulated.matched, [[1234571, 3, 1]])

    def test__process_sp(self):
        mock_runner = mock.Mock()
        mock_runner.sp.actual_sp = 12.20
        self.simulated._process_sp(1234571, mock_runner)
        self.assertEqual(self.simulated.matched, [[1234571, 12.2, 2.00]])
        self.assertTrue(self.simulated._bsp_reconciled)
        self.simulated.order.execution_complete.assert_called()

    def test__process_sp_lay(self):
        mock_runner = mock.Mock()
        mock_runner.sp.actual_sp = 12.20
        self.simulated.order.side = "LAY"
        self.simulated._process_sp(1234571, mock_runner)
        self.assertEqual(self.simulated.matched, [[1234571, 12.2, 1.96]])
        self.assertEqual(self.simulated.size_cancelled, 0.04)
        self.assertTrue(self.simulated._bsp_reconciled)
        self.simulated.order.execution_complete.assert_called()

    def test__process_sp_lay_liability(self):
        self.simulated.size_matched = 1.90
        mock_runner = mock.Mock()
        mock_runner.sp.actual_sp = 12.20
        self.simulated.order.side = "LAY"
        self.simulated._process_sp(1234571, mock_runner)
        self.assertEqual(self.simulated.matched, [])
        self.assertEqual(self.simulated.size_lapsed, 0.10)
        self.assertTrue(self.simulated._bsp_reconciled)
        self.simulated.order.execution_complete.assert_called()

    def test__process_sp_none(self):
        mock_runner = mock.Mock()
        mock_runner.sp.actual_sp = None
        self.simulated._process_sp(1234571, mock_runner)
        self.assertEqual(self.simulated.matched, [])
        self.assertFalse(self.simulated._bsp_reconciled)

    def test__process_sp_processed_semi(self):
        mock_runner = mock.Mock()
        mock_runner.sp.actual_sp = 12.20
        self.simulated.matched = [[1234571, 10.0, 1]]
        self.simulated.size_matched = 1
        self.simulated.average_price_matched = 10.0
        self.simulated._process_sp(1234572, mock_runner)
        self.assertEqual(
            self.simulated.matched, [[1234571, 10.0, 1], [1234572, 12.2, 1]]
        )
        self.assertTrue(self.simulated._bsp_reconciled)
        self.simulated.order.execution_complete.assert_called()

    def test__process_sp_limit_on_close_back(self):
        mock_limit_on_close_order = mock.Mock(price=10.0, liability=2.00)
        mock_limit_on_close_order.ORDER_TYPE = OrderTypes.LIMIT_ON_CLOSE
        self.simulated.order.order_type = mock_limit_on_close_order
        mock_runner_book = mock.Mock()
        mock_runner_book.sp.actual_sp = 69
        self.simulated._process_sp(1234573, mock_runner_book)
        self.assertEqual(self.simulated.matched, [[1234573, 69, 2.00]])
        self.assertTrue(self.simulated._bsp_reconciled)
        self.simulated.order.execution_complete.assert_called()

    def test__process_sp_limit_on_close_back_no_match(self):
        mock_limit_on_close_order = mock.Mock(price=10.0, liability=2.00)
        mock_limit_on_close_order.ORDER_TYPE = OrderTypes.LIMIT_ON_CLOSE
        self.simulated.order.order_type = mock_limit_on_close_order
        mock_runner_book = mock.Mock()
        mock_runner_book.sp.actual_sp = 8
        self.simulated._process_sp(1234574, mock_runner_book)
        self.assertEqual(self.simulated.matched, [])
        self.assertTrue(self.simulated._bsp_reconciled)
        self.simulated.order.execution_complete.assert_called()

    def test__process_sp_limit_on_close_lay(self):
        mock_limit_on_close_order = mock.Mock(price=100.0, liability=2.00)
        mock_limit_on_close_order.ORDER_TYPE = OrderTypes.LIMIT_ON_CLOSE
        self.simulated.order.order_type = mock_limit_on_close_order
        self.simulated.order.side = "LAY"
        mock_runner_book = mock.Mock()
        mock_runner_book.sp.actual_sp = 69
        self.simulated._process_sp(1234575, mock_runner_book)
        self.assertEqual(self.simulated.matched, [[1234575, 69, 0.03]])
        self.assertTrue(self.simulated._bsp_reconciled)
        self.simulated.order.execution_complete.assert_called()

    def test__process_sp_limit_on_close_lay_no_match(self):
        mock_limit_on_close_order = mock.Mock(price=60.0, liability=2.00)
        mock_limit_on_close_order.ORDER_TYPE = OrderTypes.LIMIT_ON_CLOSE
        self.simulated.order.order_type = mock_limit_on_close_order
        self.simulated.order.side = "LAY"
        mock_runner_book = mock.Mock()
        mock_runner_book.sp.actual_sp = 69
        self.simulated._process_sp(1234576, mock_runner_book)
        self.assertEqual(self.simulated.matched, [])
        self.assertTrue(self.simulated._bsp_reconciled)
        self.simulated.order.execution_complete.assert_called()

    def test__process_sp_market_on_close_back(self):
        mock_limit_on_close_order = mock.Mock(liability=2.00)
        mock_limit_on_close_order.ORDER_TYPE = OrderTypes.MARKET_ON_CLOSE
        self.simulated.order.order_type = mock_limit_on_close_order
        mock_runner_book = mock.Mock()
        mock_runner_book.sp.actual_sp = 69
        self.simulated._process_sp(1234577, mock_runner_book)
        self.assertEqual(self.simulated.matched, [[1234577, 69, 2.00]])
        self.assertTrue(self.simulated._bsp_reconciled)
        self.simulated.order.execution_complete.assert_called()

    def test__process_sp_market_on_close_lay(self):
        mock_limit_on_close_order = mock.Mock(liability=2.00)
        mock_limit_on_close_order.ORDER_TYPE = OrderTypes.MARKET_ON_CLOSE
        self.simulated.order.order_type = mock_limit_on_close_order
        self.simulated.order.side = "LAY"
        mock_runner_book = mock.Mock()
        mock_runner_book.sp.actual_sp = 69
        self.simulated._process_sp(1234578, mock_runner_book)
        self.assertEqual(self.simulated.matched, [[1234578, 69, 0.03]])
        self.assertTrue(self.simulated._bsp_reconciled)
        self.simulated.order.execution_complete.assert_called()

    @mock.patch(
        "flumine.simulation.simulatedorder.SimulatedOrder._calculate_process_traded"
    )
    def test__process_traded_back(self, mock__calculate_process_traded):
        self.simulated._process_traded(1234579, {12: 120})
        mock__calculate_process_traded.assert_called_with(1234579, 120)

    @mock.patch(
        "flumine.simulation.simulatedorder.SimulatedOrder._calculate_process_traded"
    )
    def test__process_traded_back_no(self, mock__calculate_process_traded):
        self.simulated._process_traded(1234580, {11: 120})
        mock__calculate_process_traded.assert_not_called()

    @mock.patch(
        "flumine.simulation.simulatedorder.SimulatedOrder._calculate_process_traded"
    )
    def test__process_traded_lay(self, mock__calculate_process_traded):
        self.simulated.order.side = "LAY"
        self.simulated._process_traded(1234581, {12: 120})
        mock__calculate_process_traded.assert_called_with(1234581, 120)

    @mock.patch(
        "flumine.simulation.simulatedorder.SimulatedOrder._calculate_process_traded"
    )
    def test__process_traded_lay_no(self, mock__calculate_process_traded):
        self.simulated.order.side = "LAY"
        self.simulated._process_traded(1234582, {13: 120})
        mock__calculate_process_traded.assert_not_called()

    def test__calculate_process_traded(self):
        self.assertEqual(self.simulated._calculate_process_traded(1234582, 2.00), 2)
        self.assertEqual(self.simulated._calculate_process_traded(1234583, 2.00), 2)
        self.assertEqual(self.simulated._calculate_process_traded(1234584, 2.00), 0)
        self.assertEqual(
            self.simulated.matched, [[1234582, 12, 1.00], [1234583, 12, 1.00]]
        )
        self.assertEqual(self.simulated._piq, 0)

    def test__calculate_process_traded_piq(self):
        self.simulated._piq = 2.00
        self.assertEqual(self.simulated._calculate_process_traded(1234585, 4.00), 4)
        self.assertEqual(self.simulated.matched, [])
        self.assertEqual(self.simulated._piq, 0)
        self.assertEqual(self.simulated._calculate_process_traded(1234586, 4.00), 4)
        self.assertEqual(self.simulated.matched, [[1234586, 12, 2.00]])
        self.assertEqual(self.simulated._piq, 0)
        self.assertEqual(self.simulated._calculate_process_traded(1234586, 4.00), 0)

    def test__calculate_process_traded_piq_match(self):
        self.simulated._piq = 4.00
        self.assertEqual(self.simulated._calculate_process_traded(1234585, 20.00), 12)
        self.assertEqual(self.simulated.matched, [[1234585, 12, 2.0]])
        self.assertEqual(self.simulated._piq, 0)

    def test_take_sp(self):
        self.assertFalse(self.simulated.take_sp)
        self.simulated.order.order_type.ORDER_TYPE = OrderTypes.LIMIT_ON_CLOSE
        self.assertTrue(self.simulated.take_sp)
        self.simulated.order.order_type.ORDER_TYPE = OrderTypes.MARKET_ON_CLOSE
        self.assertTrue(self.simulated.take_sp)
        self.simulated.order.order_type.ORDER_TYPE = OrderTypes.LIMIT
        self.simulated.order.order_type.persistence_type = "MARKET_ON_CLOSE"
        self.assertTrue(self.simulated.take_sp)

    def test_side(self):
        self.assertEqual(self.simulated.side, self.mock_order.side)

    def test_average_price_matched(self):
        self.assertEqual(self.simulated.average_price_matched, 0)
        self.simulated._update_matched([1234, 1, 2])
        self.assertEqual(self.simulated.average_price_matched, 1)

    def test_size_matched(self):
        self.assertEqual(self.simulated.size_matched, 0)
        self.simulated._update_matched([4321, 1, 2])
        self.assertEqual(self.simulated.size_matched, 2)

    def test__update_matched(self):
        self.simulated._update_matched([12345, 10.0, 2.64])
        self.assertEqual(self.simulated.matched, [[12345, 10.0, 2.64]])
        self.assertEqual(self.simulated.size_matched, 2.64)
        self.assertEqual(self.simulated.average_price_matched, 10.0)

    def test_size_remaining(self):
        self.assertEqual(self.simulated.size_remaining, 2)
        self.simulated._update_matched([1234, 1, 1])
        self.assertEqual(self.simulated.size_remaining, 1)

    def test_size_remaining_target(self):
        self.mock_order.order_type.size = None
        self.mock_order.order_type.bet_target_size = 2
        self.assertEqual(self.simulated.size_remaining, 2)
        self.simulated._update_matched([1234, 1, 1])
        self.assertEqual(self.simulated.size_remaining, 1)

    def test_size_remaining_multi(self):
        self.simulated._update_matched([1234, 1, 0.1])
        self.simulated.size_cancelled = 0.2
        self.simulated.size_lapsed = 0.3
        self.simulated.size_voided = 0.4
        self.assertEqual(self.simulated.size_remaining, 1)

    def test_size_remaining_non_limit(self):
        self.simulated.order.order_type.liability = 2
        self.simulated.order.order_type.ORDER_TYPE = OrderTypes.LIMIT_ON_CLOSE
        self.assertEqual(self.simulated.size_remaining, 0)

    def test_profit_back(self):
        self.assertEqual(self.simulated.profit, 0)
        self.simulated.order.runner_status = "WINNER"
        self.simulated._update_matched([1234, 10.0, 2.0])
        self.assertEqual(self.simulated.profit, 18.0)
        self.simulated.order.runner_status = "LOSER"
        self.assertEqual(self.simulated.profit, -2.0)

    def test_profit_lay(self):
        self.simulated.order.side = "LAY"
        self.assertEqual(self.simulated.profit, 0)
        self.simulated.order.runner_status = "WINNER"
        self.simulated._update_matched([1234, 10.0, 2.0])
        self.assertEqual(self.simulated.profit, -18.0)
        self.simulated.order.runner_status = "LOSER"
        self.assertEqual(self.simulated.profit, 2.0)

    def test_profit_dead_heat_back(self):
        self.simulated.order.number_of_dead_heat_winners = 2

        self.simulated.order.runner_status = "WINNER"
        self.simulated._update_matched([1234, 4.0, 10.0])
        self.assertEqual(self.simulated.profit, 10)

        self.simulated.order.runner_status = "LOSER"
        self.assertEqual(self.simulated.profit, -10)

    def test_profit_dead_heat_lay(self):
        self.simulated.order.side = "LAY"
        self.simulated.order.number_of_dead_heat_winners = 2

        self.simulated.order.runner_status = "WINNER"
        self.simulated._update_matched([1234, 5.0, 20.0])
        self.assertEqual(self.simulated.profit, -30)

        self.simulated.order.runner_status = "LOSER"
        self.assertEqual(self.simulated.profit, 20)

    def test_profit_dead_heat_four(self):
        self.simulated.order.number_of_dead_heat_winners = 4
        self.simulated.order.runner_status = "WINNER"
        self.simulated._update_matched([1234, 11.0, 50.0])
        self.assertEqual(self.simulated.profit, 87.50)

    def test_profit_ew_back(self):
        self.simulated.order.market_type = "EACH_WAY"
        self.simulated.order.each_way_divisor = 5
        self.assertEqual(self.simulated.profit, 0)
        self.simulated.order.runner_status = "WINNER"
        self.simulated._update_matched([1234, 10.0, 2.0])
        self.assertEqual(self.simulated.profit, 21.6)
        self.simulated.order.runner_status = "PLACED"
        self.assertEqual(self.simulated.profit, 1.6)
        self.simulated.order.runner_status = "LOSER"
        self.assertEqual(self.simulated.profit, -4.0)

    def test_profit_ew_lay(self):
        self.simulated.order.side = "LAY"
        self.simulated.order.market_type = "EACH_WAY"
        self.simulated.order.each_way_divisor = 5
        self.assertEqual(self.simulated.profit, 0)
        self.simulated.order.runner_status = "WINNER"
        self.simulated._update_matched([1234, 10.0, 2.0])
        self.assertEqual(self.simulated.profit, -21.6)
        self.simulated.order.runner_status = "PLACED"
        self.assertEqual(self.simulated.profit, -1.6)
        self.simulated.order.runner_status = "LOSER"
        self.assertEqual(self.simulated.profit, 4.0)

    @mock.patch(
        "flumine.simulation.simulatedorder.SimulatedOrder.size_remaining",
        new_callable=mock.PropertyMock,
    )
    @mock.patch(
        "flumine.simulation.simulatedorder.SimulatedOrder.take_sp",
        new_callable=mock.PropertyMock,
        return_value=False,
    )
    def test_status(self, mock_take_sp, mock_size_remaining):
        mock_size_remaining.return_value = 1
        self.assertEqual(self.simulated.status, "EXECUTABLE")
        mock_size_remaining.return_value = 0
        self.assertEqual(self.simulated.status, "EXECUTION_COMPLETE")

    @mock.patch(
        "flumine.simulation.simulatedorder.SimulatedOrder.take_sp", return_value=True
    )
    def test_status_sp(self, mock_take_sp):
        self.simulated._bsp_reconciled = False
        self.assertEqual(self.simulated.status, "EXECUTABLE")
        self.simulated._bsp_reconciled = True
        self.assertEqual(self.simulated.status, "EXECUTION_COMPLETE")

    @mock.patch(
        "flumine.simulation.simulatedorder.SimulatedOrder.size_remaining",
        return_value=1,
        new_callable=mock.PropertyMock,
    )
    @mock.patch(
        "flumine.simulation.simulatedorder.SimulatedOrder.take_sp",
        return_value=False,
        new_callable=mock.PropertyMock,
    )
    def test_status_limit(self, mock_take_sp, mock_size_remaining):
        self.assertEqual(self.simulated.status, "EXECUTABLE")
        mock_size_remaining.return_value = 0
        self.assertEqual(self.simulated.status, "EXECUTION_COMPLETE")

    def test_info(self):
        self.assertEqual(
            self.simulated.info,
            {
                "profit": 0,
                "piq": 0,
                "matched": [],
            },
        )

    def test_bool(self):
        self.simulated.order.client.paper_trade = False
        self.assertFalse(self.simulated)
        from flumine import config

        config.simulated = True
        self.assertTrue(self.simulated)
        config.simulated = False

    def test_bool_paper_trade(self):
        self.simulated.order.client.paper_trade = False
        self.assertFalse(self.simulated)
        self.simulated.order.client.paper_trade = True
        self.assertTrue(self.simulated)



import unittest
from unittest import mock

from flumine.markets.blotter import Blotter, PENDING_STATUS
from flumine.order.order import OrderStatus
from flumine.order.ordertype import MarketOnCloseOrder, LimitOrder, LimitOnCloseOrder


class BlotterTest(unittest.TestCase):
    def setUp(self) -> None:
        self.blotter = Blotter("1.23")

    def test_init(self):
        self.assertEqual(self.blotter.market_id, "1.23")
        self.assertFalse(self.blotter.active)
        self.assertEqual(self.blotter._orders, {})
        self.assertEqual(self.blotter._bet_id_lookup, {})
        self.assertEqual(self.blotter._live_orders, [])
        self.assertEqual(self.blotter._trades, {})
        self.assertEqual(self.blotter._strategy_orders, {})
        self.assertEqual(self.blotter._strategy_selection_orders, {})
        self.assertEqual(self.blotter._client_orders, {})
        self.assertEqual(self.blotter._client_strategy_orders, {})
        self.assertEqual(
            PENDING_STATUS,
            [
                OrderStatus.PENDING,
                OrderStatus.VIOLATION,
                OrderStatus.EXPIRED,
            ],
        )

    def test_get_order_bet_id(self):
        self.assertIsNone(self.blotter.get_order_bet_id("123"))
        mock_order = mock.Mock(selection_id=2, handicap=3, bet_id="123")
        self.blotter["456"] = mock_order
        self.assertEqual(self.blotter.get_order_bet_id("123"), mock_order)

    def test_strategy_orders(self):
        mock_order_one = mock.Mock(
            selection_id=2, handicap=3, status=OrderStatus.EXECUTABLE, size_matched=1
        )
        mock_order_one.trade.strategy = 69
        self.blotter["12345"] = mock_order_one
        mock_order_two = mock.Mock(
            selection_id=2,
            handicap=3,
            status=OrderStatus.EXECUTION_COMPLETE,
            size_matched=0,
        )
        mock_order_two.trade.strategy = 69
        self.blotter["12345"] = mock_order_two
        mock_order_three = mock.Mock(
            selection_id=2,
            handicap=3,
            status=OrderStatus.EXECUTION_COMPLETE,
            size_matched=1,
        )
        mock_order_three.trade.strategy = 69
        self.blotter["12345"] = mock_order_three
        self.assertEqual(self.blotter.strategy_orders(12), [])
        self.assertEqual(
            self.blotter.strategy_orders(69),
            [mock_order_one, mock_order_two, mock_order_three],
        )
        self.assertEqual(
            self.blotter.strategy_orders(
                69, order_status=[OrderStatus.EXECUTION_COMPLETE]
            ),
            [mock_order_two, mock_order_three],
        )
        self.assertEqual(
            self.blotter.strategy_orders(
                69, order_status=[OrderStatus.EXECUTION_COMPLETE], matched_only=True
            ),
            [mock_order_three],
        )

    def test_strategy_selection_orders(self):
        mock_order_one = mock.Mock(
            selection_id=2, handicap=3, status=OrderStatus.EXECUTABLE, size_matched=1
        )
        mock_order_one.trade.strategy = 69
        self.blotter["12345"] = mock_order_one
        mock_order_two = mock.Mock(
            selection_id=2,
            handicap=3,
            status=OrderStatus.EXECUTION_COMPLETE,
            size_matched=0,
        )
        mock_order_two.trade.strategy = 69
        self.blotter["12345"] = mock_order_two
        mock_order_three = mock.Mock(
            selection_id=2,
            handicap=3,
            status=OrderStatus.EXECUTION_COMPLETE,
            size_matched=1,
        )
        mock_order_three.trade.strategy = 69
        self.blotter["12345"] = mock_order_three
        self.assertEqual(self.blotter.strategy_selection_orders(12, 2, 3), [])
        self.assertEqual(
            self.blotter.strategy_selection_orders(69, 2, 3),
            [mock_order_one, mock_order_two, mock_order_three],
        )
        self.assertEqual(
            self.blotter.strategy_selection_orders(
                69, 2, 3, order_status=[OrderStatus.EXECUTION_COMPLETE]
            ),
            [mock_order_two, mock_order_three],
        )
        self.assertEqual(
            self.blotter.strategy_selection_orders(
                69,
                2,
                3,
                order_status=[OrderStatus.EXECUTION_COMPLETE],
                matched_only=True,
            ),
            [mock_order_three],
        )

    def test_client_orders(self):
        mock_client_one = mock.Mock()
        mock_order_one = mock.Mock(
            client=mock_client_one,
            selection_id=2,
            handicap=3,
            status=OrderStatus.EXECUTABLE,
            size_matched=1,
        )
        mock_order_one.trade.strategy = 69
        self.blotter["12345"] = mock_order_one
        mock_order_two = mock.Mock(
            client=mock_client_one,
            selection_id=2,
            handicap=3,
            status=OrderStatus.EXECUTION_COMPLETE,
            size_matched=0,
        )
        mock_order_two.trade.strategy = 69
        self.blotter["12345"] = mock_order_two
        mock_order_three = mock.Mock(
            client=mock_client_one,
            selection_id=2,
            handicap=3,
            status=OrderStatus.EXECUTION_COMPLETE,
            size_matched=1,
        )
        mock_order_three.trade.strategy = 69
        self.blotter["12345"] = mock_order_three
        self.assertEqual(self.blotter.client_strategy_orders(mock_client_one, 12), [])
        self.assertEqual(
            self.blotter.client_strategy_orders(mock_client_one, 69),
            [mock_order_one, mock_order_two, mock_order_three],
        )
        self.assertEqual(
            self.blotter.client_strategy_orders(
                mock_client_one, 69, order_status=[OrderStatus.EXECUTION_COMPLETE]
            ),
            [mock_order_two, mock_order_three],
        )
        self.assertEqual(
            self.blotter.client_strategy_orders(
                mock_client_one,
                69,
                order_status=[OrderStatus.EXECUTION_COMPLETE],
                matched_only=True,
            ),
            [mock_order_three],
        )

    def test_client_strategy_orders(self):
        mock_client_one = mock.Mock()
        mock_order_one = mock.Mock(
            client=mock_client_one,
            selection_id=2,
            handicap=3,
            status=OrderStatus.EXECUTABLE,
            size_matched=1,
        )
        self.blotter["12345"] = mock_order_one
        mock_order_two = mock.Mock(
            client=mock_client_one,
            selection_id=2,
            handicap=3,
            status=OrderStatus.EXECUTION_COMPLETE,
            size_matched=0,
        )
        self.blotter["12345"] = mock_order_two
        mock_order_three = mock.Mock(
            client=mock_client_one,
            selection_id=2,
            handicap=3,
            status=OrderStatus.EXECUTION_COMPLETE,
            size_matched=1,
        )
        self.blotter["12345"] = mock_order_three
        self.assertEqual(self.blotter.client_orders(12), [])
        self.assertEqual(
            self.blotter.client_orders(mock_client_one),
            [mock_order_one, mock_order_two, mock_order_three],
        )
        self.assertEqual(
            self.blotter.client_orders(
                mock_client_one, order_status=[OrderStatus.EXECUTION_COMPLETE]
            ),
            [mock_order_two, mock_order_three],
        )
        self.assertEqual(
            self.blotter.client_orders(
                mock_client_one,
                order_status=[OrderStatus.EXECUTION_COMPLETE],
                matched_only=True,
            ),
            [mock_order_three],
        )

    def test_live_orders(self):
        self.assertEqual(list(self.blotter.live_orders), [])
        mock_order = mock.Mock(complete=False)
        self.blotter._live_orders = [mock_order]
        self.assertEqual(list(self.blotter.live_orders), [mock_order])

    def test_has_live_orders(self):
        self.assertFalse(self.blotter.has_live_orders)
        self.blotter._live_orders = [mock.Mock()]
        self.assertTrue(self.blotter.has_live_orders)

    def test_process_closed_market(self):
        mock_market_book = mock.Mock(number_of_winners=1)
        mock_runner = mock.Mock(selection_id=123, handicap=0.0)
        mock_market_book.runners = [mock_runner]
        mock_order = mock.Mock(selection_id=123, handicap=0.0)
        self.blotter._orders = {"12345": mock_order}
        self.blotter.process_closed_market(mock_market_book)
        self.assertEqual(mock_order.runner_status, mock_runner.status)
        self.assertEqual(
            mock_order.market_type, mock_market_book.market_definition.market_type
        )
        self.assertEqual(
            mock_order.each_way_divisor,
            mock_market_book.market_definition.each_way_divisor,
        )

    def test_process_cleared_orders(self):
        mock_cleared_orders = mock.Mock()
        mock_cleared_orders.orders = []
        self.assertEqual(self.blotter.process_cleared_orders(mock_cleared_orders), [])

    def test_selection_exposure(self):
        """
        Check that selection_exposure returns the absolute worse loss
        """

        def get_exposures(strategy, lookup):
            if strategy == "strategy" and lookup == (1, 2, 3):
                return {
                    "worst_possible_profit_on_win": -1.0,
                    "worst_possible_profit_on_lose": -2.0,
                }

        self.blotter.get_exposures = mock.Mock(side_effect=get_exposures)

        result = self.blotter.selection_exposure("strategy", (1, 2, 3))

        self.assertEqual(2.0, result)

    def test_selection_exposure2(self):
        """
        Check that selection_exposure returns zero if there is no risk of loss.
        """

        def get_exposures(strategy, lookup):
            if strategy == "strategy" and lookup == (1, 2, 3):
                return {
                    "worst_possible_profit_on_win": 0.0,
                    "worst_possible_profit_on_lose": 1.0,
                }

        self.blotter.get_exposures = mock.Mock(side_effect=get_exposures)

        result = self.blotter.selection_exposure("strategy", (1, 2, 3))

        self.assertEqual(0.0, result)

    def test_get_exposures(self):
        mock_strategy = mock.Mock()
        mock_trade = mock.Mock(strategy=mock_strategy)
        mock_order = mock.Mock(
            trade=mock_trade,
            lookup=(self.blotter.market_id, 123, 0),
            selection_id=123,
            handicap=0,
            side="BACK",
            average_price_matched=5.6,
            size_matched=2.0,
            size_remaining=0.0,
            order_type=LimitOrder(price=5.6, size=2.0),
        )
        self.blotter["12345"] = mock_order
        self.assertEqual(
            self.blotter.get_exposures(mock_strategy, mock_order.lookup),
            {
                "matched_profit_if_lose": -2.0,
                "matched_profit_if_win": 9.2,
                "worst_possible_profit_on_lose": -2.0,
                "worst_possible_profit_on_win": 9.2,
                "worst_potential_unmatched_profit_if_lose": 0.0,
                "worst_potential_unmatched_profit_if_win": 0.0,
            },
        )

    def test_get_exposures_with_exclusion(self):
        mock_strategy = mock.Mock()
        mock_trade = mock.Mock(strategy=mock_strategy)
        mock_order = mock.Mock(
            trade=mock_trade,
            lookup=(self.blotter.market_id, 123, 0),
            selection_id=123,
            handicap=0,
            side="BACK",
            average_price_matched=5.6,
            size_matched=2.0,
            size_remaining=0.0,
            order_type=LimitOrder(price=5.6, size=2.0),
        )
        mock_order_excluded = mock.Mock(
            trade=mock_trade,
            lookup=(self.blotter.market_id, 123, 0),
            selection_id=123,
            handicap=0,
            side="BACK",
            average_price_matched=5.6,
            size_matched=2.0,
            size_remaining=0.0,
            order_type=LimitOrder(price=5.6, size=2.0),
        )
        self.blotter["12345"] = mock_order
        self.blotter["67890"] = mock_order_excluded
        self.assertEqual(
            self.blotter.get_exposures(
                mock_strategy, mock_order.lookup, exclusion=mock_order_excluded
            ),
            {
                "matched_profit_if_lose": -2.0,
                "matched_profit_if_win": 9.2,
                "worst_possible_profit_on_lose": -2.0,
                "worst_possible_profit_on_win": 9.2,
                "worst_potential_unmatched_profit_if_lose": 0.0,
                "worst_potential_unmatched_profit_if_win": 0.0,
            },
        )

    def test_get_exposures_value_error(self):
        mock_strategy = mock.Mock()
        mock_trade = mock.Mock(strategy=mock_strategy)
        mock_order = mock.Mock(
            trade=mock_trade,
            lookup=(self.blotter.market_id, 123, 0),
            selection_id=123,
            handicap=0,
            side="BACK",
            average_price_matched=5.6,
            size_matched=2.0,
            size_remaining=0.0,
            order_type=mock.Mock(ORDER_TYPE="INVALID"),
        )
        self.blotter["12345"] = mock_order

        with self.assertRaises(ValueError) as e:
            self.blotter.get_exposures(mock_strategy, mock_order.lookup)

        self.assertEqual("Unexpected order type: INVALID", e.exception.args[0])

    def test_get_exposures_with_price_none(self):
        """
        Check that get_exposures works if order.order_type.price is None.
        If order.order_type.price is None, the controls will flag the order as a violation
        and it won't be set to the exchange, so there won't be any exposure and we can ignore it.
        """
        mock_strategy = mock.Mock()
        mock_trade = mock.Mock(strategy=mock_strategy)
        lookup = (self.blotter.market_id, 123, 0)
        mock_order1 = mock.Mock(
            trade=mock_trade,
            lookup=lookup,
            selection_id=123,
            handicap=0,
            side="BACK",
            average_price_matched=5.6,
            size_matched=2.0,
            size_remaining=0.0,
            order_type=LimitOrder(price=5.6, size=2.0),
        )
        mock_order2 = mock.Mock(
            trade=mock_trade,
            lookup=lookup,
            selection_id=123,
            handicap=0,
            side="LAY",
            average_price_matched=5.6,
            size_matched=0.0,
            size_remaining=2.0,
            order_type=LimitOrder(price=None, size=2.0),
        )
        self.blotter["12345"] = mock_order1
        self.blotter["23456"] = mock_order2
        self.assertEqual(
            self.blotter.get_exposures(mock_strategy, lookup),
            {
                "matched_profit_if_lose": -2.0,
                "matched_profit_if_win": 9.2,
                "worst_possible_profit_on_lose": -2.0,
                "worst_possible_profit_on_win": 9.2,
                "worst_potential_unmatched_profit_if_lose": 0.0,
                "worst_potential_unmatched_profit_if_win": 0.0,
            },
        )

    def test_get_exposures_no_match(self):
        mock_strategy = mock.Mock()
        mock_trade = mock.Mock(strategy=mock_strategy)
        mock_order = mock.Mock(
            trade=mock_trade,
            lookup=(self.blotter.market_id, 123, 0),
            selection_id=123,
            handicap=0,
            side="BACK",
            average_price_matched=5.6,
            size_matched=0.0,
            size_remaining=0.0,
            order_type=LimitOrder(price=5.6, size=2.0),
        )
        self.blotter["12345"] = mock_order
        self.assertEqual(
            self.blotter.get_exposures(mock_strategy, mock_order.lookup),
            {
                "matched_profit_if_lose": 0.0,
                "matched_profit_if_win": 0.0,
                "worst_possible_profit_on_lose": 0.0,
                "worst_possible_profit_on_win": 0.0,
                "worst_potential_unmatched_profit_if_lose": 0.0,
                "worst_potential_unmatched_profit_if_win": 0.0,
            },
        )

    def test_get_exposures_from_unmatched_back(self):
        mock_strategy = mock.Mock()
        mock_trade = mock.Mock(strategy=mock_strategy)
        mock_order = mock.Mock(
            trade=mock_trade,
            lookup=(self.blotter.market_id, 123, 0),
            selection_id=123,
            handicap=0,
            side="BACK",
            average_price_matched=5.6,
            size_matched=2.0,
            size_remaining=2.0,
            order_type=LimitOrder(price=6, size=4.0),
            status=OrderStatus.EXECUTABLE,
            complete=False,
        )
        self.blotter["12345"] = mock_order
        # On the win side, we have 2.0 * (5.6-1.0) = 9.2
        # On the lose side, we have -2.0-2.0=-4.0
        self.assertEqual(
            self.blotter.get_exposures(mock_strategy, mock_order.lookup),
            {
                "matched_profit_if_lose": -2.0,
                "matched_profit_if_win": 9.2,
                "worst_possible_profit_on_lose": -4.0,
                "worst_possible_profit_on_win": 9.2,
                "worst_potential_unmatched_profit_if_lose": -2.0,
                "worst_potential_unmatched_profit_if_win": 0,
            },
        )

    def test_get_exposures_from_unmatched_lay(self):
        mock_strategy = mock.Mock()
        mock_trade = mock.Mock(strategy=mock_strategy)
        mock_order = mock.Mock(
            trade=mock_trade,
            lookup=(self.blotter.market_id, 123, 0),
            selection_id=123,
            handicap=0,
            side="LAY",
            average_price_matched=5.6,
            size_matched=2.0,
            size_remaining=2.0,
            order_type=LimitOrder(price=6, size=4.0),
            status=OrderStatus.EXECUTABLE,
            complete=False,
        )
        self.blotter["12345"] = mock_order
        # On the win side, we have -2.0 * (5.6-1.0) -2.0 * (6.0-1.0) = -19.2
        # On the lose side, we have 2.0 from size_matched
        self.assertEqual(
            self.blotter.get_exposures(mock_strategy, mock_order.lookup),
            {
                "matched_profit_if_lose": 2.0,
                "matched_profit_if_win": -9.2,
                "worst_possible_profit_on_lose": 2.0,
                "worst_possible_profit_on_win": -19.2,
                "worst_potential_unmatched_profit_if_lose": 0,
                "worst_potential_unmatched_profit_if_win": -10.0,
            },
        )

    def test_get_exposures_from_market_on_close_back(self):
        mock_strategy = mock.Mock()
        mock_trade = mock.Mock(strategy=mock_strategy)
        mock_order = mock.Mock(
            trade=mock_trade,
            lookup=(self.blotter.market_id, 123, 0),
            selection_id=123,
            handicap=0,
            side="BACK",
            order_type=MarketOnCloseOrder(liability=10.0),
        )
        self.blotter["12345"] = mock_order
        self.assertEqual(
            self.blotter.get_exposures(mock_strategy, mock_order.lookup),
            {
                "matched_profit_if_lose": 0.0,
                "matched_profit_if_win": 0.0,
                "worst_possible_profit_on_lose": -10.0,
                "worst_possible_profit_on_win": 0.0,
                "worst_potential_unmatched_profit_if_lose": 0.0,
                "worst_potential_unmatched_profit_if_win": 0.0,
            },
        )

    def test_get_exposures_from_market_on_close_lay(self):
        mock_strategy = mock.Mock()
        mock_trade = mock.Mock(strategy=mock_strategy)
        mock_order = mock.Mock(
            trade=mock_trade,
            lookup=(self.blotter.market_id, 123, 0),
            selection_id=123,
            handicap=0,
            side="LAY",
            order_type=MarketOnCloseOrder(liability=10.0),
        )
        self.blotter["12345"] = mock_order
        self.assertEqual(
            self.blotter.get_exposures(mock_strategy, mock_order.lookup),
            {
                "matched_profit_if_lose": 0.0,
                "matched_profit_if_win": 0.0,
                "worst_possible_profit_on_lose": 0.0,
                "worst_possible_profit_on_win": -10.0,
                "worst_potential_unmatched_profit_if_lose": 0.0,
                "worst_potential_unmatched_profit_if_win": 0.0,
            },
        )

    def test_get_exposures_from_limit_on_close_lay(self):
        mock_strategy = mock.Mock()
        mock_trade = mock.Mock(strategy=mock_strategy)
        mock_order = mock.Mock(
            trade=mock_trade,
            lookup=(self.blotter.market_id, 123, 0),
            selection_id=123,
            handicap=0,
            side="LAY",
            order_type=LimitOnCloseOrder(price=1.01, liability=10.0),
        )
        self.blotter["12345"] = mock_order
        self.assertEqual(
            self.blotter.get_exposures(mock_strategy, mock_order.lookup),
            {
                "matched_profit_if_lose": 0.0,
                "matched_profit_if_win": 0.0,
                "worst_possible_profit_on_lose": 0.0,
                "worst_possible_profit_on_win": -10.0,
                "worst_potential_unmatched_profit_if_lose": 0.0,
                "worst_potential_unmatched_profit_if_win": 0.0,
            },
        )

    def test_get_exposures_voided(self):
        mock_strategy = mock.Mock()
        mock_trade = mock.Mock(strategy=mock_strategy)
        mock_order = mock.Mock(
            trade=mock_trade,
            lookup=(self.blotter.market_id, 123, 0),
            selection_id=123,
            handicap=0,
            side="BACK",
            order_type=LimitOrder(price=5, size=10.0),
            status=OrderStatus.VIOLATION,
        )
        self.blotter["12345"] = mock_order
        self.assertEqual(
            self.blotter.get_exposures(mock_strategy, mock_order.lookup),
            {
                "matched_profit_if_lose": 0.0,
                "matched_profit_if_win": 0.0,
                "worst_possible_profit_on_lose": 0.0,
                "worst_possible_profit_on_win": 0.0,
                "worst_potential_unmatched_profit_if_lose": 0.0,
                "worst_potential_unmatched_profit_if_win": 0.0,
            },
        )

    def test_market_position(self):
        mock_strategy = mock.Mock()
        mock_trade = mock.Mock(strategy=mock_strategy)
        order_data = [
            # (order_id, selection_id, side, average_price_matched, size_matched, size_remaining, order_type, complete)
            (1001, 123, "BACK", 5.6, 2.0, 0.0, LimitOrder(price=5.6, size=2.0), True),
            (1002, 123, "LAY", 5.2, 3.0, 0.0, LimitOrder(price=5.2, size=3.0), True),
            (1003, 234, "LAY", 4.8, 4.0, 1.0, LimitOrder(price=4.8, size=5.0), True),
            (1004, 456, "BACK", None, 0.0, 0.0, LimitOrder(price=5.6, size=2.0), False),
            (1005, 678, "BACK", None, 0, 0, MarketOnCloseOrder(liability=6), False),
            (
                1006,
                678,
                "LAY",
                None,
                0,
                0,
                LimitOnCloseOrder(price=1.01, liability=10),
                False,
            ),
        ]
        for order in order_data:
            self.blotter[order[0]] = mock.Mock(
                trade=mock_trade,
                lookup=(self.blotter.market_id, order[1], 0),
                selection_id=order[1],
                handicap=0,
                side=order[2],
                average_price_matched=order[3],
                size_matched=order[4],
                size_remaining=order[5],
                order_type=order[6],
                complete=order[7],
            )

        mock_market_book = mock.Mock(number_of_active_runners=6, number_of_winners=1)
        self.assertEqual(  # single winner
            self.blotter.market_exposure(mock_strategy, mock_market_book), -20.2
        )
        mock_market_book = mock.Mock(number_of_active_runners=6, number_of_winners=2)
        self.assertEqual(  # muliple winners
            self.blotter.market_exposure(mock_strategy, mock_market_book), -24.6
        )
        mock_market_book = mock.Mock(number_of_active_runners=20, number_of_winners=7)
        self.assertEqual(  # num winners > num runners traded
            self.blotter.market_exposure(mock_strategy, mock_market_book), -28.6
        )

    def test_greened_market_position(self):
        mock_strategy = mock.Mock()
        mock_market_book = mock.Mock(number_of_active_runners=6, number_of_winners=1)
        mock_trade = mock.Mock(strategy=mock_strategy)
        orders = [
            # (order_id, selection_id, side, average_price_matched, size_matched, size_remaining, order_type, complete)
            (1001, 123, "BACK", 5.6, 2.0, 0.0, LimitOrder(price=5.6, size=2.0), True),
            (1002, 123, "LAY", 5.2, 2.1, 0.0, LimitOrder(price=5.2, size=2.1), True),
            (1003, 234, "BACK", 4.8, 4.0, 0.0, LimitOrder(price=4.8, size=4.0), True),
            (1004, 234, "LAY", 4, 4.2, 0.0, LimitOrder(price=4, size=4.2), True),
            (1005, 345, "BACK", 10, 2.0, 0.0, LimitOrder(price=10, size=2.0), True),
            (1006, 345, "LAY", 8, 2.2, 0.0, LimitOrder(price=8, size=2.2), True),
        ]
        for order in orders:
            self.blotter[order[0]] = mock.Mock(
                trade=mock_trade,
                lookup=(self.blotter.market_id, order[1], 0),
                selection_id=order[1],
                handicap=0,
                side=order[2],
                average_price_matched=order[3],
                size_matched=order[4],
                size_remaining=order[5],
                order_type=order[6],
                complete=order[7],
            )
        self.assertEqual(  # single winner
            self.blotter.market_exposure(mock_strategy, mock_market_book), 0.5
        )

    def test_complete_order(self):
        self.blotter._live_orders = ["test"]
        self.blotter.complete_order("test")

    def test_has_trade(self):
        self.assertFalse(self.blotter.has_trade("123"))
        self.blotter._trades["123"].append(1)
        self.assertTrue(self.blotter.has_trade("123"))

    def test__contains(self):
        self.blotter._orders = {"123": "test"}
        self.assertIn("123", self.blotter)
        self.assertNotIn("321", self.blotter)

    def test__setitem(self):
        mock_client = mock.Mock()
        mock_order = mock.Mock(
            selection_id=2, handicap=3, bet_id="456", client=mock_client
        )
        self.blotter["123"] = mock_order
        self.assertTrue(self.blotter.active)
        self.assertEqual(self.blotter._orders, {"123": mock_order})
        self.assertEqual(self.blotter._bet_id_lookup, {"456": mock_order})
        self.assertEqual(self.blotter._live_orders, [mock_order])
        self.assertEqual(self.blotter._trades, {mock_order.trade.id: [mock_order]})
        self.assertEqual(
            self.blotter._strategy_orders, {mock_order.trade.strategy: [mock_order]}
        )
        self.assertEqual(
            self.blotter._strategy_selection_orders,
            {(mock_order.trade.strategy, 2, 3): [mock_order]},
        )
        self.assertEqual(
            self.blotter._client_orders,
            {mock_client: [mock_order]},
        )
        self.assertEqual(
            self.blotter._client_strategy_orders,
            {(mock_client, mock_order.trade.strategy): [mock_order]},
        )

    def test__getitem(self):
        self.blotter._orders = {"12345": "test", "54321": "test2"}
        self.assertEqual(self.blotter["12345"], "test")
        self.assertEqual(self.blotter["54321"], "test2")

    def test__len(self):
        self.blotter._orders = {"12345": "test", "54321": "test"}
        self.assertEqual(len(self.blotter), 2)
